# Copyright 2021 - 2025 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Loading schemapack and datapack definitions.

Warning: This is an internal part of the library and might change without notice.
"""

from pathlib import Path

import pydantic

from schemapack._internals.utils import transient_directory_change
from schemapack.exceptions import DataPackSpecError, SchemaPackSpecError
from schemapack.spec.datapack import DataPack
from schemapack.spec.schemapack import SchemaPack
from schemapack.utils import read_json_or_yaml_mapping


def load_schemapack(path: Path):
    """Load a schemapack definition from a file."""
    schemapack_dict = read_json_or_yaml_mapping(path)

    with transient_directory_change(path.parent):
        try:
            return SchemaPack.model_validate(schemapack_dict)
        except pydantic.ValidationError as error:
            raise SchemaPackSpecError(
                message=str(error), details=error.errors()
            ) from error


def load_datapack(path: Path):
    """Load a datapack definition from a file."""
    datapack_dict = read_json_or_yaml_mapping(path)

    try:
        return DataPack.model_validate(datapack_dict)
    except pydantic.ValidationError as error:
        raise DataPackSpecError(message=str(error), details=error.errors()) from error



# Copyright 2021 - 2025 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""A validation plugin."""

from collections.abc import Mapping

from schemapack._internals.validation.base import ClassValidationPlugin
from schemapack.exceptions import ValidationPluginError
from schemapack.spec.custom_types import ResourceId
from schemapack.spec.datapack import DataPack, Resource
from schemapack.spec.schemapack import ClassDefinition


class MissingMandatoryOriginValidationPlugin(ClassValidationPlugin):
    """A class-scoped validation plugin validating that every target resource is
    referenced by at least one origin resource for relations that are mandatory at the
    origin end.

    This only applies to schemapacks classes with relations that are mandatory at the
    origin end.
    """

    @staticmethod
    def does_apply(*, class_: ClassDefinition) -> bool:
        """A classmethod to check whether this validation plugin is relevant for the
        given schemapack.

        Returns: True if this plugin is relevant for the given class definition.
        """
        return any(relation.mandatory.origin for relation in class_.relations.values())

    def __init__(self, *, class_: ClassDefinition):
        """This plugin is configured with one specific class definition of a schemapack."""
        self._relations_of_interest = {
            relation_name: relation
            for relation_name, relation in class_.relations.items()
            if relation.mandatory.origin
        }

    def validate(
        self, *, class_resources: Mapping[ResourceId, Resource], datapack: DataPack
    ):
        """Validate all resources of a specific class. The entire datapack is provided
        for resolving relations to other classes.

        Raises:
            schemapack.exceptions.ValidationPluginError: If validation fails.
        """
        not_referenced_target_id_by_relation: dict[str, set[str]] = {}

        for relation_name, relation in self._relations_of_interest.items():
            try:
                referenced_target_ids = {
                    target_id
                    for resource in class_resources.values()
                    for target_id in resource.relations[
                        relation_name
                    ].get_target_resources_as_set()
                }
            except KeyError:
                referenced_target_ids = set()

            all_possible_target_ids = set(
                datapack.resources.get(relation.targetClass, {})
            )

            not_referenced_target_ids = all_possible_target_ids.difference(
                referenced_target_ids
            )
            if not_referenced_target_ids:
                not_referenced_target_id_by_relation[relation_name] = (
                    not_referenced_target_ids
                )

        if any(not_referenced_target_id_by_relation.values()):
            raise ValidationPluginError(
                type_="MissingMandatoryOriginError",
                message=(
                    "For the following relations that are mandatory on the origin end"
                    + " the specified targets are not referenced by any origin:"
                    + f"{not_referenced_target_id_by_relation}"
                ),
                details={
                    "not_referenced_target_by_relation": not_referenced_target_id_by_relation
                },
            )



# Copyright 2021 - 2025 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Test dumping a schemapack."""

import json
from pathlib import Path

import pytest
import ruamel.yaml

from schemapack import (
    dump_schemapack,
    dumps_datapack,
    dumps_schemapack,
    load_datapack,
    load_schemapack,
)
from schemapack._internals.utils import read_json_or_yaml_mapping
from tests.fixtures.examples import VALID_DATAPACK_PATHS, VALID_SCHEMAPACK_PATHS
from tests.fixtures.utils import (
    assert_formatted_string,
    loads_json_or_yaml_mapping,
)

yaml = ruamel.yaml.YAML(typ="rt")


@pytest.mark.parametrize(
    "yaml_format", [True, False], ids=["yaml_format", "json_format"]
)
def test_dumps_datapack(yaml_format: bool):
    """Tests using the dumps_datapack function."""
    datapack_path = VALID_DATAPACK_PATHS["simple_relations.simple_resources"]
    datapack = load_datapack(datapack_path)
    expected_dict = read_json_or_yaml_mapping(datapack_path)

    observed_str = dumps_datapack(datapack, yaml_format=yaml_format)
    assert_formatted_string(observed_str, json_format=not yaml_format)

    observed_dict = loads_json_or_yaml_mapping(observed_str)
    assert observed_dict == expected_dict


@pytest.mark.parametrize(
    "yaml_format", [True, False], ids=["yaml_format", "json_format"]
)
def test_dumps_schemapack(yaml_format: bool):
    """Tests using the dumps_schemapack function to dump a schemapack as a
    condensed representation to string.
    """
    schemapack = load_schemapack(VALID_SCHEMAPACK_PATHS["simple_relations"])
    expected_dict = read_json_or_yaml_mapping(
        VALID_SCHEMAPACK_PATHS["simple_relations_condensed"]
    )

    observed_str = dumps_schemapack(schemapack, yaml_format=yaml_format)
    assert_formatted_string(observed_str, json_format=not yaml_format)

    observed_dict = loads_json_or_yaml_mapping(observed_str)
    assert observed_dict == expected_dict


@pytest.mark.parametrize(
    "yaml_format", [True, False], ids=["yaml_format", "json_format"]
)
def test_dump_condensed_schemapack(yaml_format: bool, tmp_path: Path):
    """Tests using the dump_schemapack function to dump a schemapack as a
    condensed representation to file.
    """
    input_schemapack = load_schemapack(VALID_SCHEMAPACK_PATHS["simple_relations"])
    expected_dict = yaml.load(VALID_SCHEMAPACK_PATHS["simple_relations_condensed"])
    output_path = tmp_path / "output.schemapack.yaml"

    dump_schemapack(input_schemapack, path=output_path, yaml_format=yaml_format)
    if yaml_format:
        observed_dict = yaml.load(output_path)
    else:
        with open(output_path, encoding="utf-8") as file:
            observed_dict = json.load(file)

    assert observed_dict == expected_dict


def test_dump_not_condensed_schemapack(tmp_path: Path):
    """Tests using the dump_schemapack function to dump a schemapack as a representation
    to file with content schemas being written to dedicated files.
    """
    input_path = VALID_SCHEMAPACK_PATHS["simple_relations"]
    input_schemapack = load_schemapack(input_path)
    expected_dict = yaml.load(input_path)
    schemapack_dir = tmp_path / "schemapack" / "valid"
    output_path = schemapack_dir / "output.schemapack.yaml"
    rel_content_schema_path = Path("../../content_schemas/")

    schemapack_dir.mkdir(parents=True)

    dump_schemapack(
        input_schemapack,
        path=output_path,
        condensed=False,
        content_schema_dir=rel_content_schema_path,
    )

    # check schemapack file itself:
    observed_dict = yaml.load(output_path)
    assert observed_dict == expected_dict

    # check content schema files:
    for class_name, class_ in input_schemapack.classes.items():
        content_schema_path = (
            schemapack_dir / rel_content_schema_path / f"{class_name}.schema.json"
        )
        with open(content_schema_path, encoding="utf-8") as file:
            observed_content_schema = json.load(file)
        expected_content_schema = json.loads(class_.model_dump_json())["content"]
        assert observed_content_schema == expected_content_schema



# Copyright 2021 - 2025 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Models representing the datapack spec.

Warning: This is an internal part of the library and might change without notice.
"""

import typing
from collections import Counter
from collections.abc import Iterable, Mapping
from typing import Annotated, Any, Literal, Self, TypeAlias

from arcticfreeze import FrozenDict, freeze
from pydantic import (
    BeforeValidator,
    Field,
    WrapSerializer,
    field_serializer,
    model_validator,
)
from pydantic_core import PydanticCustomError

from schemapack._internals.spec.base import _FrozenNoExtraBaseModel
from schemapack._internals.spec.custom_types import (
    ClassName,
    RelationPropertyName,
    ResourceId,
)
from schemapack._internals.utils import thaw

SupportedDataPackVersions = Literal["3.0.0", "3.1.0", "4.0.0"]
SUPPORTED_DATA_PACK_VERSIONS = typing.get_args(SupportedDataPackVersions)


def validate_duplicate_target_ids(iterable: Iterable) -> Any:
    """Checks that the given iterable of target IDs does not contain duplicates. If it
    does, a PydanticCustomError with name "DuplicateTargetIdError" is raised. Otherwise,
    the given iterable is returned.
    """
    if isinstance(iterable, frozenset):
        return iterable

    try:
        target_id_list = list(iterable)
    except TypeError as error:
        raise PydanticCustomError(
            "TargetIdsParsingError",
            "The provided object is not iterable.",
            {"iterable": iterable},
        ) from error

    counter = Counter(target_id_list)
    duplicates = {k for k, v in counter.items() if v > 1}

    if duplicates:
        raise PydanticCustomError(
            "DuplicateTargetIdError",
            "The given sequence of target ids contain duplicates: {duplicates}",
            {"duplicates": duplicates},
        )

    return iterable


ResourceIdSet: TypeAlias = Annotated[
    frozenset[str],
    # Upon serialization, assert that the provided sequence does not contain duplicates:
    BeforeValidator(validate_duplicate_target_ids),
    # Upon serialization, produce predictablily sorted lists:
    WrapSerializer(lambda v, next_: sorted(next_(v))),
]

ContentPropertyValue: TypeAlias = Annotated[
    Any,
    # the value of a content property is deeply frozen:
    BeforeValidator(lambda obj: freeze(obj, by_superclass=True)),
]


class ResourceRelation(_FrozenNoExtraBaseModel):
    """A model defining relations of a resource to other resources."""

    targetClass: ClassName = Field(  # noqa: N815 - following JSON conventions
        ...,
        description=("The name of the target class of the relation."),
    )
    targetResources: ResourceId | ResourceIdSet | None = Field(  # noqa: N815 - following JSON conventions
        ...,
        description=(
            "Provides the ID(s) of target resources of the targetClass. Depending on"
            + " the corresponding schemapack definition, this field could be one of the"
            + " following types:"
            + " (1) a id of a single target resource (schemapack assumes that"
            + " multiple.target is False),"
            + " (2) None (schemapack assumes that multiple.target and mandatory.target"
            + " are both False),"
            + " (3) a set of ids of target resources (schemapack assumes that"
            + " multiple.target is True),"
            + " (4) an empty set (schemapack assumes that multiple.target is True and"
            + " mandatory.target is False)."
        ),
    )

    def get_target_resources_as_set(self) -> frozenset[ResourceId]:
        """Get target resources as a set independent of the multiplicity or
        mandatoriness. This is even the case if the actual value in the relations dict
        is a single
        string (translated into a list of length one) or None (translated into an
        empty set). If do_not_raise is True, the method will return an empty set
        even if the relation name does not exist in the relations dict.
        """
        if self.targetResources is None:
            return frozenset()
        if isinstance(self.targetResources, frozenset):
            return self.targetResources
        return frozenset({self.targetResources})


class Resource(_FrozenNoExtraBaseModel):
    """A model defining content and relations of a resource
    of a specific class.
    """

    content: FrozenDict[str, ContentPropertyValue] = Field(
        ...,
        description=(
            "The content of the resource that complies with the content schema defined"
            + " in the corresponding schemapack."
        ),
    )

    relations: FrozenDict[RelationPropertyName, ResourceRelation] = Field(
        default_factory=FrozenDict,
        description=(
            "A dictionary containing the relations of the resource to other resources."
            + " Each key correspond to the name of a relation property. Each value"
            + "  contains the target class and target resource(s) of the relation."
        ),
    )


class DataPack(_FrozenNoExtraBaseModel):
    """A model for describing a schemapack definition."""

    datapack: SupportedDataPackVersions = Field(
        ...,
        description=(
            "Has two purposes: (1) it clearly identifies a YAML/JSON document as"
            + " a datapack definition and (2) it specifies the used version of the"
            + " datapack specification."
        ),
    )

    resources: FrozenDict[ClassName, FrozenDict[ResourceId, Resource]] = Field(
        ...,
        description=(
            "A nested dictionary containing resources per class name (keys on the first"
            + " level) and resource ID (keys on the second level). Each class defined"
            + " in the schemapack must be present even if no resources are defined for"
            + " it in this datapack."
        ),
    )

    rootResource: str | None = Field(  # noqa: N815 - following JSON conventions
        None,
        description=(
            "Defines the id of the resource of the class defined in `className`"
            + " that should act as root. This means"
            + " that, in addition to the root resource itself, the datapack must only"
            + " contain resources that are direct or indirect (dependencies of"
            + " dependencies) of the root resource."
        ),
    )

    rootClass: ClassName | None = Field(  # noqa: N815 - following JSON conventions
        None,
        description=(
            "Defines the class name of the resource that should act as root. This means"
            + " that, in addition to the root resource itself, the datapack must only"
            + " contain resources that are direct or indirect (dependencies of"
            + " dependencies) of the root resource."
        ),
    )

    @field_serializer("resources", mode="plain")
    def serialize_resources(self, value: Any) -> dict[str, object]:
        return thaw(value)

    @model_validator(mode="before")
    @classmethod
    def check_root_duality(cls, value: Mapping) -> Mapping | None:
        """Ensures that both 'rootClass' and 'rootResource' are either present or absent."""
        missing = [key for key in ("rootClass", "rootResource") if not value.get(key)]

        if len(missing) == 1:
            raise PydanticCustomError(
                "DatapackRootDualityError",
                "Invalid DataPack due to missing field '{missing}'.",
                {"missing": missing[0]},
            )

        return value

    @model_validator(mode="after")
    def validate_root_class(self) -> Self:
        """Checks if 'rootClass' exists in DataPack resources when specified."""
        if self.rootClass and self.rootClass not in self.resources:
            raise PydanticCustomError(
                "UnknownRootClassError",
                "The class '{root_class}'"
                + " given as root class does not exist among the DataPack resources",
                {"root_class": self.rootClass},
            )
        return self

    @model_validator(mode="after")
    def validate_root_resource(self) -> Self:
        """Checks if 'rootResource' exists in DataPack resources when specified."""
        if self.rootClass and self.rootResource not in self.resources[self.rootClass]:
            raise PydanticCustomError(
                "UnkownRootResourceError",
                "The specified root resource with ID '{root_resource}' of class "
                + " '{root_class}' does not exist.",
                {"root_resource": self.rootResource, "root_class": self.rootClass},
            )
        return self

    @model_validator(mode="after")
    def validate_relation_classes(self) -> Self:
        """Checks if the `targetClass` of relations defined in a datapack exists
        within datapack resources.
        """
        non_found_classes: dict[str, str] = {}  # target_class -> relation_name

        for resources in self.resources.values():
            for resource in resources.values():
                for relation_name, relation in resource.relations.items():
                    if relation.targetClass not in self.resources:
                        non_found_classes[relation.targetClass] = relation_name

        if non_found_classes:
            raise PydanticCustomError(
                "TargetClassNotFoundError",
                "Did not find the target class of the following relations (relation"
                + " names): {non_found_classes}",
                {
                    "non_found_classes": ", ".join(
                        f"'{target_class}' ('{relation_name}')"
                        for target_class, relation_name in non_found_classes.items()
                    )
                },
            )
        return self

    @model_validator(mode="after")
    def validate_relation_resources(self) -> Self:
        """Checks if the `targetResources` of relations defined in a datapack exists
        within its corresponding `targetClass`.
        """
        non_found_target_ids: dict[str, str] = {}  # target_id -> relation_name

        for resources in self.resources.values():
            for resource in resources.values():
                for relation_name, relation in resource.relations.items():
                    target_ids = relation.get_target_resources_as_set()
                    for target_id in target_ids:
                        if target_id not in self.resources.get(
                            relation.targetClass, set()
                        ):
                            non_found_target_ids[target_id] = relation_name

        if non_found_target_ids:
            raise PydanticCustomError(
                "TargetIdNotFoundError",
                "Did not find a target resource for the following ID(s) (relation"
                + " names): {non_found_target_ids}",
                {
                    "non_found_target_ids": ", ".join(
                        f"'{target_id}' ('{relation_name}')"
                        for target_id, relation_name in non_found_target_ids.items()
                    )
                },
            )
        return self



# Copyright 2021 - 2025 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""A validation plugin."""

import json
from collections.abc import Mapping
from typing import Any

import jsonschema.exceptions
import jsonschema.protocols
import jsonschema.validators

from schemapack._internals.validation.base import ResourceValidationPlugin
from schemapack.exceptions import ValidationPluginError
from schemapack.spec.custom_types import ResourceId
from schemapack.spec.datapack import DataPack, Resource
from schemapack.spec.schemapack import ClassDefinition


def _get_json_schema_validator(
    schema: Mapping[str, Any],
) -> jsonschema.protocols.Validator:
    """Get a JSON Schema validator for the given schema.
    It is assumed that the schema has already been checked for validity against the
    JSON Schema specs.
    """
    cls: type[jsonschema.protocols.Validator] = jsonschema.validators.validator_for(
        schema
    )
    return cls(schema)


class ContentSchemaValidationPlugin(ResourceValidationPlugin):
    """A resource-scoped validation plugin validating the content of one resource
    against the content JSON Schema defined in the corresponding schemapack.
    """

    @staticmethod
    def does_apply(*, class_: ClassDefinition) -> bool:
        """A classmethod to check whether this validation plugin is relevant for the
        given schemapack.

        Returns: True if this plugin is relevant for the given class definition.
        """
        # Is always relevant since all resources must have a content schema:
        return True

    def __init__(self, *, class_: ClassDefinition):
        """This plugin is configured with one specific class definition of a schemapack."""
        self._json_schema_validator = _get_json_schema_validator(class_.content)

    def validate(
        self, *, resource: Resource, resource_id: ResourceId, datapack: DataPack
    ):
        """Validates a specific resource of the defined class. The entire datapack is
        provided for resolving relations to resources of other classes.

        Raises:
            schemapack.exceptions.ValidationPluginError: If validation fails.
        """
        json_compatible_content = json.loads(resource.model_dump_json())["content"]
        try:
            self._json_schema_validator.validate(json_compatible_content)
        except jsonschema.exceptions.ValidationError as error:
            raise ValidationPluginError(
                type_="ContentValidationError", message=error.message
            ) from error



# Copyright 2021 - 2025 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Integrate rooted datapacks into nested json objects."""

from collections import defaultdict
from collections.abc import Mapping
from dataclasses import dataclass
from typing import Any, TypeAlias

from schemapack._internals.exceptions import InvalidEmbeddingProfileError
from schemapack._internals.spec.custom_types import EmbeddingProfile
from schemapack._internals.spec.datapack import Resource
from schemapack._internals.spec.schemapack import ClassDefinition
from schemapack._internals.utils import thaw
from schemapack.exceptions import CircularRelationError, ValidationAssumptionError
from schemapack.spec.custom_types import ClassName, ResourceId
from schemapack.spec.datapack import DataPack
from schemapack.spec.schemapack import SchemaPack

JsonObjectCompatible: TypeAlias = dict[str, object]


@dataclass
class DenormalizationContext:
    """Context for denormalization to group repeated arguments."""

    datapack: DataPack
    schemapack: SchemaPack
    embedding_profile: Mapping[str, Any] | None
    resource_blacklist: dict[ClassName, set[ResourceId]]


def denormalize(
    *,
    datapack: DataPack,
    schemapack: SchemaPack,
    embedding_profile: EmbeddingProfile = None,
    _resource_blacklist: Mapping[ClassName, set[ResourceId]] | None = None,
    _alt_root_class_name: ClassName | None = None,
    _alt_root_resource_id: ResourceId | None = None,
) -> JsonObjectCompatible:
    """Integrate a rooted datapack into a nested json object-compatible data structure.
    It is assumed that the provided datapack has already been validated against the
    provided schemapack.

    Args:
        datapack:
            The datapack to be denormalized. Must be rooted.
        schemapack:
            The schemapack to be used for looking up the classes of relations.
        embedding_profile:
            An optional profile defining which relations should be embedded and which
            should not. If None, all relations will be embedded. If a relation is not
            present in the profile, it will be embedded by default.
        _resource_blacklist:
            An optional blacklist of resources (a mapping with resource ids as values
            and class names as keys) that cause an error to be raised if they are
            encountered when assembling the json compatible object. This is only used
            internally to prevent circular relations.
        _alt_root_class_name:
            An optional alternative root class name to be used instead of the class
            name specified as root in the datapack. This is only used internally for
            recursive calls.
        _alt_root_resource_id:
            An optional alternative root resource id to be used instead of the resource
            id specified as root in the datapack. This is only used internally for
            recursive calls.

    Raises:
        ValueError:
            If the datapack is not rooted.
        schemapack.exceptions.ValidationAssumptionError:
            If the datapack is not valid against the schemapack.
        schemapack.exceptions.CircularRelationError:
            If a circular relation is detected.
    """
    if not datapack.rootResource:
        raise ValueError("Datapack must be rooted.")

    if not datapack.rootClass:
        raise ValueError("Datapack must have a root class.")

    root_class_name = _alt_root_class_name or datapack.rootClass
    root_resource_id = _alt_root_resource_id or datapack.rootResource

    resource_blacklist = _initialize_blacklist(
        embedding_profile, root_class_name, root_resource_id, _resource_blacklist
    )

    root_resource = _get_root_resource(datapack, root_class_name, root_resource_id)

    root_class_definition = _get_class_definition(schemapack, root_class_name)

    denormalized_object: JsonObjectCompatible = {
        root_class_definition.id.propertyName: root_resource_id,
        **thaw(root_resource.content),
    }

    for relation_name, resource_relations in root_resource.relations.items():
        target_ids = resource_relations.targetResources
        target_class_name = resource_relations.targetClass

        # decide whether to embed the relation 'relation_name'
        should_embed = _should_embed(embedding_profile, relation_name)

        if should_embed is False:
            denormalized_object[relation_name] = (
                [*sorted(target_ids)]
                if isinstance(target_ids, frozenset)
                else target_ids
                if isinstance(target_ids, str)
                else []
            )
            continue

        next_embedding_profile = _get_next_embedding_profile(
            embedding_profile, relation_name
        )
        denormalization_context = DenormalizationContext(
            datapack=datapack,
            schemapack=schemapack,
            embedding_profile=next_embedding_profile,
            resource_blacklist=resource_blacklist,
        )
        denormalized_object[relation_name] = _process_recursion(
            denormalization_context,
            target_class_name,
            target_ids,
        )

    return denormalized_object


def _initialize_blacklist(
    embedding_profile: EmbeddingProfile,
    root_class_name: ClassName,
    root_resource_id: ResourceId,
    _resource_blacklist: Mapping[ClassName, set[ResourceId]] | None,
) -> dict[ClassName, set[ResourceId]]:
    """Function to initialize the resource blacklist."""
    resource_blacklist: dict[ClassName, set[ResourceId]] = defaultdict(set)
    if not embedding_profile:
        resource_blacklist[root_class_name].add(root_resource_id)

    if _resource_blacklist:
        for class_name, resource_ids in _resource_blacklist.items():
            resource_blacklist[class_name].update(resource_ids)
    return resource_blacklist


def _get_root_resource(
    datapack: DataPack, root_class_name: ClassName, root_resource_id: ResourceId
) -> Resource:
    """Function to get the root resource from the datapack."""
    root_class_resources = datapack.resources.get(root_class_name)
    if not root_class_resources:
        raise ValidationAssumptionError(context="root class lookup")

    root_resource = root_class_resources.get(root_resource_id)
    if not root_resource:
        raise ValidationAssumptionError(context="root resource lookup")

    return root_resource


def _get_class_definition(
    schemapack: SchemaPack, root_class_name: ClassName
) -> ClassDefinition:
    """Function to get the class definition from the schemapack."""
    root_class_definition = schemapack.classes.get(root_class_name)
    if not root_class_definition:
        raise RuntimeError(
            "This is a bug and should not happen. It should be caught by the schemapack"
            + " spec validation."
        )
    return root_class_definition


def _should_embed(embedding_profile: EmbeddingProfile, relation_name: str) -> bool:
    """Function to decide whether to embed a relation based on the embedding profile.
    When a relation_name is not found in the profile, it returns True by default.
    If the profile is None, it returns True for all relations.
    If the profile is empty, it returns True for all relations.
    """
    if not embedding_profile:
        return True

    value = embedding_profile.get(relation_name)
    if value is None:
        return True

    if isinstance(value, bool):
        return value
    if isinstance(value, dict):
        return True

    raise InvalidEmbeddingProfileError(
        f"Invalid embedding profile: expected bool or dict for '{relation_name}', got {type(value).__name__}"
    )


def _get_next_embedding_profile(
    embedding_profile: EmbeddingProfile, relation_name: str
) -> EmbeddingProfile:
    """Function to get the next embedding profile for a relation."""
    if not embedding_profile:
        return None
    value = embedding_profile.get(relation_name)
    if isinstance(value, dict):
        return value
    return None


def _process_recursion(
    context: DenormalizationContext,
    class_name: ClassName,
    resource_ids: frozenset[ResourceId] | ResourceId | None,
) -> list[JsonObjectCompatible] | JsonObjectCompatible | None:
    """Function to process the recursion for denormalization."""
    if isinstance(resource_ids, frozenset):
        return [
            _recursive_denormalize(
                context,
                class_name,
                target_id,
            )
            for target_id in sorted(resource_ids)
        ]
    elif isinstance(resource_ids, str):
        return _recursive_denormalize(
            context,
            class_name,
            resource_ids,
        )
    else:
        return None


def _recursive_denormalize(
    context: DenormalizationContext,
    class_name: ClassName,
    resource_id: ResourceId,
):
    """Function used in recursively denormalize a resource by calling the function 'denormalize'."""
    if resource_id in context.resource_blacklist.get(class_name, set()):
        raise CircularRelationError(
            f"Cannot perform denormalization of datapack with circular relations. "
            f"The circular relation involved the resource with id {resource_id} of class {class_name}."
        )

    return denormalize(
        datapack=context.datapack,
        schemapack=context.schemapack,
        embedding_profile=context.embedding_profile,
        _resource_blacklist=context.resource_blacklist,
        _alt_root_class_name=class_name,
        _alt_root_resource_id=resource_id,
    )



# Copyright 2021 - 2025 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""A validation plugin."""

from schemapack._internals.validation.base import GlobalValidationPlugin
from schemapack.exceptions import ValidationPluginError
from schemapack.spec.datapack import DataPack
from schemapack.spec.schemapack import SchemaPack


class UnexpectedRootValidationPlugin(GlobalValidationPlugin):
    """A global-scoped validation plugin validating that a datapack has no root resource
    and root class defined.
    This plugin is only relevant if the schemapack has no root class defined.
    """

    @staticmethod
    def does_apply(*, schemapack: SchemaPack) -> bool:
        """A classmethod to check whether this validation plugin is relevant for the
        given schemapack.

        Returns: True if this plugin is relevant for the given class definition.
        """
        return not bool(schemapack.rootClass)

    def __init__(self, *, schemapack: SchemaPack):
        """This plugin is configured with the entire schemapack."""
        # there is nothing to do

    def validate(self, *, datapack: DataPack):
        """Validate the entire datapack.

        Raises:
            schemapack.exceptions.ValidationPluginError: If validation fails.
        """
        defined = []
        if datapack.rootResource:
            defined.append("root resource")
        if datapack.rootClass:
            defined.append("root class")

        if defined:
            raise ValidationPluginError(
                type_="UnexpectedRootDefinitionError",
                message=(
                    "The schemapack has no root class defined but the datapack "
                    f"specifies a {' and '.join(defined)}."
                ),
            )



# Copyright 2021 - 2025 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Models representing the schemapack spec.

Warning: This is an internal part of the library and might change without notice.
"""

import typing
from collections.abc import Mapping
from pathlib import Path
from typing import Any, Literal, cast

from arcticfreeze import FrozenDict, freeze
from immutabledict import immutabledict
from pydantic import (
    Field,
    field_validator,
    model_validator,
)
from pydantic_core import PydanticCustomError

from schemapack._internals.spec.base import _FrozenNoExtraBaseModel
from schemapack._internals.utils import JsonSchemaError, assert_valid_json_schema
from schemapack.exceptions import ParsingError
from schemapack.spec.custom_types import (
    ClassName,
    ContentPropertyName,
    IdPropertyName,
    RelationPropertyName,
)
from schemapack.utils import read_json_or_yaml_mapping

SupportedSchemaPackVersions = Literal["3.0.0", "3.1.0", "4.0.0"]
SUPPORTED_SCHEMA_PACK_VERSIONS = typing.get_args(SupportedSchemaPackVersions)


def validate_object_json_schema(value: Mapping[str, Any]):
    """Check if the given dict represents a valid JSON Schema for object types.

    Raises:
        PydanticCustomError:
            If the value is not a valid JSON Schema for object types.
    """
    try:
        assert_valid_json_schema(value)
    except JsonSchemaError as error:
        raise PydanticCustomError(
            "InvalidContentSchemaError",
            "The content schema is not a valid JSON schema: {error_message}",
            {"error_message": str(error)},
        ) from error

    if value.get("type") != "object":
        raise PydanticCustomError(
            "InvalidContentSchemaError",
            "The content schema must be an object.",
        )

    return value


class MandatoryRelationSpec(_FrozenNoExtraBaseModel):
    """A model for describing the modality of a relation. It describes the minimum
    number of instances the origin and the target end must contribute to the relation.
    """

    origin: bool = Field(
        ...,
        description=(
            "If true, the origin must participate in the relation."
            + " I.e. every instance of the target class must be connected to at least"
            + " one instance of the origin class trough this relation."
            + " If false, the participation of the origin is optional."
            + " I.e. an instance of the target class may be connected to zero or more"
            + " instances of the origin class trough this relation."
        ),
    )
    target: bool = Field(
        ...,
        description=(
            "If true, the target must participate in the relation."
            + " I.e. every instance of the origin class must be connected to at least"
            + " one instance of the target class trough this relation."
            + " If false, the participation of the target is optional."
            + " I.e. an instance of the origin class may be connected to zero or more"
            + " instances of the target class trough this relation."
        ),
    )


class MultipleRelationSpec(_FrozenNoExtraBaseModel):
    """A model for describing the cardinality of a relation. It describes the maximum
    number of instances the origin and the target end may contribute to the relation.

    For instance, if the origin is `True` and target is `False`, the origin may
    contribute multiple instances to the relation, while the target may at most
    contribute a single instance to the relation. This is equivalent to a 'many-to-one'
    """

    origin: bool = Field(
        ...,
        description=(
            "If true, the origin may contribute multiple instances to the relation."
            + " This is equivalent to a 'many-to-*' cardinality."
            + " If false, the origin may at most contribute a single instance to the"
            + " relation. This is equivalent to a 'one-to-*' cardinality."
        ),
    )
    target: bool = Field(
        ...,
        description=(
            "If true, the target may contribute multiple instances to the relation."
            + " This is equivalent to a '*-to-many' cardinality."
            + " If false, the target may at most contribute a single instance to the"
            + " relation. This is equivalent to a '*-to-one' cardinality."
        ),
    )


class ClassRelation(_FrozenNoExtraBaseModel):
    """A model for describing a schemapack relation definition."""

    description: str | None = Field(
        None,
        description="A description of the relation.",
    )
    targetClass: str = Field(  # noqa: N815 - align with the schemapack naming scheme
        ...,
        description="The name of the target class.",
    )
    mandatory: MandatoryRelationSpec = Field(
        ...,
        description=(
            "The modality of the relation. It describes the minimum number of instances"
            + " the origin and the target end must contribute to the relation."
        ),
    )
    multiple: MultipleRelationSpec = Field(
        ...,
        description=(
            "The cardinality of the relation. It describes the maximum number of"
            + " instances the origin and the target end may contribute to the relation."
            + " For instance, if the origin is `True` and target is `False`, the origin"
            + " may contribute multiple instances to the relation, while the target may"
            + " at most contribute a single instance to the relation."
            + " This is equivalent to a 'many-to-one' cardinality."
        ),
    )


class IDSpec(_FrozenNoExtraBaseModel):
    """A model for describing the ID property of a class definition."""

    propertyName: IdPropertyName = Field(  # noqa: N815 - align with the schemapack naming scheme
        ...,
        description=(
            "The name of the ID property. It must not collide with content or relations"
            + " properties."
            + "This name e.g. relavant for specifying the ID property in a"
            + " denormalized representation."
        ),
    )
    description: str | None = Field(
        None,
        description="A description of the ID property.",
    )


class ClassDefinition(_FrozenNoExtraBaseModel):
    """A model for describing a schemapack class definition."""

    description: str | None = Field(
        None,
        description=("A description of the class definition."),
    )
    id: IDSpec = Field(
        ...,
        description="The ID property of the class definition.",
    )
    content: FrozenDict[str, Any] = Field(
        ...,
        description=(
            "The content schema of the class definition. It must be a valid JSON schema"
            + " object for object types. You may also provide the path to a JSON or YAML"
            + " file containing the schema. It will be automatically loaded."
        ),
    )
    relations: FrozenDict[RelationPropertyName, ClassRelation] = Field(
        immutabledict(),
        description=(
            "A mapping of relation names to relation definitions. Relation names"
            + " should use snake_case and may only contain alphanumeric characters and"
            + " underscores. They must start with a letter."
        ),
    )  # pyright: ignore

    def get_content_properties(self) -> frozenset[ContentPropertyName]:
        """Returns a set of the content properties."""
        return frozenset(self.content.get("properties", {}))

    @field_validator("content", mode="before")
    @classmethod
    def load_and_validate_content_schema(
        cls, value: str | Path | Mapping
    ) -> FrozenDict:
        """A validator function for content schemas that:
        - loads a JSON or YAML file if a path is provided
        - checks if the value is a valid JSON schema object
        - freezes the dict representation of the schema
        """
        if isinstance(value, str):
            # assume that the string is a path to a JSON or YAML file
            value = Path(value)

        if isinstance(value, Path):
            if not value.is_file():
                absolute_path = value.absolute().resolve()
                raise PydanticCustomError(
                    "ContentSchemaNotFoundError",
                    "Content schema path does not exist or is not a file."
                    + " Absolute path: {absolute_path}",
                    {
                        "absolute_path": absolute_path,
                    },
                )

            try:
                value = read_json_or_yaml_mapping(value)
            except ParsingError as error:
                raise PydanticCustomError(
                    "InvalidContentSchemaError",
                    "Content schema at the specified path could not be parsed as"
                    + " valid JSON or YAML.",
                ) from error

        if not isinstance(value, Mapping):
            raise PydanticCustomError(
                "InvalidContentSchemaError",
                "Expected a Mapping or a path (pathlib.Path or str) to a JSON or"
                + " YAML file.",
            )

        try:
            assert_valid_json_schema(value)
        except JsonSchemaError as error:
            raise PydanticCustomError(
                "InvalidContentSchemaError",
                "The content schema is not a valid JSON schema: {error_message}",
                {"error_message": str(error)},
            ) from error

        if value.get("type") != "object":
            raise PydanticCustomError(
                "InvalidContentSchemaError", "The content schema must be an object."
            )

        return cast(FrozenDict, freeze(value, by_superclass=True))

    @field_validator("relations", mode="after")
    @classmethod
    def relation_name_validator(
        cls, v: FrozenDict[str, ClassRelation]
    ) -> FrozenDict[str, ClassRelation]:
        """Validate relation names."""
        invalid_relation_names = [
            relation_name for relation_name in v if not relation_name.isidentifier()
        ]

        if invalid_relation_names:
            raise PydanticCustomError(
                "InvalidRelationNameError",
                "Relation names may only contain alphanumeric characters and"
                + " underscores. They must not start with a number."
                + " Got {number} invalid names: {invalid_relation_names}",
                {
                    "number": len(invalid_relation_names),
                    "invalid_relation_names": invalid_relation_names,
                },
            )

        return v

    @model_validator(mode="after")
    def relation_content_property_collisions(self) -> "ClassDefinition":
        """Check for collisions between relations and content properties."""
        content_properties = self.get_content_properties()
        collisions = content_properties.intersection(set(self.relations))

        if collisions:
            raise PydanticCustomError(
                "RelationsContentPropertyCollisionError",
                "The following properties occur both in the content and the"
                + " relations: {collisions}",
                {
                    "number": len(collisions),
                    "collisions": collisions,
                },
            )

        return self

    @model_validator(mode="after")
    def id_content_property_collisions(self) -> "ClassDefinition":
        """Check for collisions between the id property and content properties."""
        if self.id.propertyName in self.get_content_properties():
            raise PydanticCustomError(
                "IdContentPropertyCollisionError",
                "The id property '{id_property}' also occurs in the content.",
                {
                    "id_property": self.id.propertyName,
                },
            )

        return self

    @model_validator(mode="after")
    def id_relations_property_collisions(self) -> "ClassDefinition":
        """Check for collisions between the id property and relations properties."""
        if self.id.propertyName in self.relations:
            raise PydanticCustomError(
                "IdRelationsPropertyCollisionError",
                "The id property '{id_property}' also occurs in the relations.",
                {
                    "id_property": self.id.propertyName,
                },
            )

        return self


class SchemaPack(_FrozenNoExtraBaseModel):
    """A model for describing a schemapack definition."""

    schemapack: SupportedSchemaPackVersions = Field(
        ...,
        description=(
            "Has two purposes: (1) it clearly identifies a YAML/JSON document as"
            + " a schemapack definition and (2) it specifies the used version of the"
            + " schemapack specification."
        ),
    )
    description: str | None = Field(
        None,
        description=("A description of the schemapack definition."),
    )
    classes: FrozenDict[ClassName, ClassDefinition] = Field(
        ...,
        description=(
            "A mapping of class names to class definitions. Class names should use"
            + " PascalCase."
        ),
        min_length=1,
    )

    rootClass: ClassName | None = Field(  # noqa: N815 - following JSON conventions
        None,
        description=(
            "Optionally, define the name of a class that should acting as the root of"
            + " the schemapack."
            + " Corresponding datapacks must define a root resource of this class."
            + "If not specified , i.e. set to None (the default), the datapack must not"
            + " specify a root resource."
        ),
    )

    @model_validator(mode="before")
    @classmethod
    def check_schemapack_field_exists(cls, data: Any) -> Any:
        """Checks that the schemapack field exists and points to a supported version
        before doing anything else. Please note, this validation only takes place if the
        data is passed as Mapping. However, the data can be pretty much anything.
        """
        if isinstance(data, Mapping):
            if "schemapack" not in data:
                raise PydanticCustomError(
                    "MissingSchemaPackVersionError",
                    "Missing a `schemapack` field. Are you sure you have passed a"
                    + " schemapack definition?",
                )

            if data["schemapack"] not in SUPPORTED_SCHEMA_PACK_VERSIONS:
                raise PydanticCustomError(
                    "UnsupportedSchemaPackVersionError",
                    "Unsupported schemapack version '{current_version}'."
                    " Supported versions are: {supported_versions}",
                    {
                        "current_version": data["schemapack"],
                        "supported_versions": SUPPORTED_SCHEMA_PACK_VERSIONS,
                    },
                )

        return data

    @field_validator("classes", mode="after")
    @classmethod
    def class_name_validator(
        cls, v: FrozenDict[str, ClassDefinition]
    ) -> FrozenDict[str, ClassDefinition]:
        """Validates class names."""
        invalid_class_names = [
            class_name for class_name in v if not class_name.isidentifier()
        ]

        if invalid_class_names:
            raise PydanticCustomError(
                "InvalidClassNameError",
                "Class names may only contain alphanumeric characters and"
                + " underscores. They must not start with a number."
                + " Got {number} invalid names: {invalid_class_names}",
                {
                    "number": len(invalid_class_names),
                    "invalid_class_names": invalid_class_names,
                },
            )

        return v

    @model_validator(mode="after")
    def relation_to_class_validation(self) -> "SchemaPack":
        """Validate that all relations point to existing classes."""
        # store invalid relations as a set of strings ({class_name}.{relation_name}):
        invalid_relations: set[str] = set()

        for class_name, class_definition in self.classes.items():
            for relation_name, relation in class_definition.relations.items():
                if relation.targetClass not in self.classes:
                    invalid_relations.add(f"{class_name}.{relation_name}")

        if invalid_relations:
            raise PydanticCustomError(
                "RelationClassNotFoundError",
                "Found {number} relation(s) that point to non-existing classes:"
                + " {invalid_relations}",
                {
                    "number": len(invalid_relations),
                    "invalid_relations": invalid_relations,
                },
            )

        return self

    @model_validator(mode="after")
    def root_in_classes(self) -> "SchemaPack":
        """Check that the specified root exists in the defined classes."""
        if self.rootClass and self.rootClass not in self.classes:
            raise PydanticCustomError(
                "RootClassNotFoundError",
                "The specified root class '{self.rootClass}' does not exist.",
                {
                    "rootClass": self.rootClass,
                },
            )

        return self



# Copyright 2021 - 2025 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for dumping schemapacks."""

import json
from pathlib import Path
from typing import Any

from schemapack._internals.spec.datapack import DataPack
from schemapack._internals.spec.schemapack import SchemaPack
from schemapack._internals.utils import (
    dumps_model,
    model_to_serializable_dict,
    write_dict,
)


def get_content_schema_path(*, class_name: str, content_schema_dir: Path) -> Path:
    """Get the path to a content schema file in the provided directory for a class with
    the provided name.
    """
    return content_schema_dir / f"{class_name}.schema.json"


def set_content_schema_paths(
    *, schemapack_dict: dict[str, Any], content_schema_dir: Path
) -> dict[str, Any]:
    """Sets the content schema paths in the provided schemapack dictionary.

    Returns:
        A copy of the provided schemapack dictionary with the content schema paths set.
    """
    modified_classes = {
        class_name: {
            **class_,
            "content": str(
                get_content_schema_path(
                    class_name=class_name, content_schema_dir=content_schema_dir
                )
            ),
        }
        for class_name, class_ in schemapack_dict["classes"].items()
    }

    return {**schemapack_dict, "classes": modified_classes}


def write_content_schemas(
    *, schemapack: SchemaPack, content_schema_dir: Path, relative_to: Path
) -> None:
    """Writes the content schemas of individual classes defined in the provided
    schemapack as separate files to the provided directory `content_schema_dir`
    which is relative to `relative_to`.
    """
    abs_content_schema_dir = relative_to / content_schema_dir
    abs_content_schema_dir.mkdir(parents=True, exist_ok=True)

    for class_name, class_ in schemapack.classes.items():
        content_schema_path = get_content_schema_path(
            class_name=class_name, content_schema_dir=abs_content_schema_dir
        )
        class_dict = json.loads(class_.model_dump_json())
        with open(content_schema_path, "w", encoding="utf-8") as file:
            json.dump(class_dict["content"], file)


def dumps_schemapack(
    schemapack: SchemaPack,
    *,
    yaml_format: bool = True,
) -> str:
    """Dumps a condensed version of the provided schemapack as a JSON or YAML-formatted
    string.

    Args:
        schemapack:
            The schemapack to dump.
        yaml_format:
            Whether to dump as YAML (`True`) or JSON (`False`).
    """
    return dumps_model(schemapack, yaml_format=yaml_format)


def dumps_datapack(
    datapack: DataPack,
    *,
    yaml_format: bool = True,
) -> str:
    """Dumps the provided datapack as a JSON or YAML-formatted string.

    Args:
        datapack:
            The datapack to dump.
        yaml_format:
            Whether to dump as YAML (`True`) or JSON (`False`).
    """
    return dumps_model(datapack, yaml_format=yaml_format)


def dump_schemapack(
    schemapack: SchemaPack,
    *,
    path: Path,
    condensed: bool = True,
    content_schema_dir: Path = Path("content_schemas"),
    yaml_format: bool = True,
) -> None:
    """Dumps the provided schemapack as a JSON or YAML file to the provided path.

    Args:
        schemapack:
            The schemapack to dump.
        path:
            The file path to write the schemapack to. The parent directory has to exist.
            If the file already exists, it will be overwritten.
        condensed:
            Whether to dump a condensed version (i.e. with content schemas embedded) of
            the provided schemapack.
        content_schema_dir:
            The path to the directory used to output the content schemas of individual
            classes defined in the provided schemapack as separate files. The path is
            relative to the `path` of the schemapack. If it does not exist, it will be
            created. The content schema files are always in JSON format and named
            according to the class name '{class_name}.schema.json'.
            Please note, this is only relevant if `condensed` is set to `False`.
        yaml_format:
            Whether to dump as YAML (`True`) or JSON (`False`).

    Raises:
        FileNotFoundError:
            If the parent directory of the provided `path` does not exist.
    """
    parent_dir = path.parent
    if not parent_dir.exists():
        raise FileNotFoundError(f"The parent directory of '{path}' does not exist.")

    schemapack_dict = model_to_serializable_dict(schemapack)

    if not condensed:
        schemapack_dict = set_content_schema_paths(
            schemapack_dict=schemapack_dict, content_schema_dir=content_schema_dir
        )
        write_content_schemas(
            schemapack=schemapack,
            content_schema_dir=content_schema_dir,
            relative_to=parent_dir,
        )

    write_dict(schemapack_dict, path=path, yaml_format=yaml_format)



# Copyright 2021 - 2025 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for running multiple validation plugins on a datapack with respect to a
schemapack.
"""

from typing import overload

from schemapack._internals.validation.base import (
    ClassValidationPlugin,
    GlobalValidationPlugin,
    ResourceValidationPlugin,
)
from schemapack._internals.validation.default import (
    DEFAULT_CLASS_PLUGIN_REGISTRY,
    DEFAULT_GLOBAL_PLUGIN_REGISTRY,
    DEFAULT_RESOURCE_PLUGIN_REGISTRY,
)
from schemapack.exceptions import (
    ValidationError,
    ValidationErrorRecord,
    ValidationPluginError,
)
from schemapack.spec.datapack import DataPack
from schemapack.spec.schemapack import SchemaPack


def _create_global_plugins(
    *, schemapack: SchemaPack, plugin_classes: list[type[GlobalValidationPlugin]]
):
    """Create instances of all the provided global plugins that apply to the given
    schemapack.
    """
    return [
        cls(schemapack=schemapack)
        for cls in plugin_classes
        if cls.does_apply(schemapack=schemapack)
    ]


@overload
def _create_plugins_by_class(
    *, schemapack: SchemaPack, plugin_classes: list[type[ClassValidationPlugin]]
) -> dict[str, list[ClassValidationPlugin]]: ...


@overload
def _create_plugins_by_class(
    *, schemapack: SchemaPack, plugin_classes: list[type[ResourceValidationPlugin]]
) -> dict[str, list[ResourceValidationPlugin]]: ...


def _create_plugins_by_class(
    *,
    schemapack: SchemaPack,
    plugin_classes: list[type[ClassValidationPlugin]]
    | list[type[ResourceValidationPlugin]]
    | list[type[ClassValidationPlugin] | type[ResourceValidationPlugin]],
) -> (
    dict[str, list[ClassValidationPlugin]]
    | dict[str, list[ResourceValidationPlugin]]
    | dict[str, list[ResourceValidationPlugin | ClassValidationPlugin]]
):
    """Create instances of the provided plugins for each class (if relevant) in
    the given schemapack.

    Returns:
        A dictionary mapping class names to the corresponding resource plugin instances.
    """
    return {
        class_name: [
            cls(class_=class_)
            for cls in plugin_classes
            if cls.does_apply(class_=class_)
        ]
        for class_name, class_ in schemapack.classes.items()
    }


def _plugin_error_to_record(
    error: ValidationPluginError,
    *,
    subject_class: str | None = None,
    subject_resource: str | None = None,
) -> ValidationErrorRecord:
    """Convert a ValidationPluginError to a ValidationErrorRecord."""
    return ValidationErrorRecord(
        subject_class=subject_class,
        subject_resource=subject_resource,
        type=error.type_,
        message=error.message,
        details=error.details,
    )


def _run_global_plugins(
    *, datapack: DataPack, plugins: list[GlobalValidationPlugin]
) -> list[ValidationErrorRecord]:
    """Run all the given global plugins on the given datapack."""
    records: list[ValidationErrorRecord] = []

    for plugin in plugins:
        try:
            plugin.validate(datapack=datapack)
        except ValidationPluginError as error:
            records.append(_plugin_error_to_record(error))

    return records


def _run_class_plugins(
    *, datapack: DataPack, plugins: dict[str, list[ClassValidationPlugin]]
) -> list[ValidationErrorRecord]:
    """Run all class plugins on all classes of the given datapack."""
    records: list[ValidationErrorRecord] = []

    for class_name, class_resources in datapack.resources.items():
        for plugin in plugins.get(class_name, []):
            try:
                plugin.validate(class_resources=class_resources, datapack=datapack)
            except ValidationPluginError as error:
                records.append(_plugin_error_to_record(error, subject_class=class_name))

    return records


def _run_resource_plugins(
    *, datapack: DataPack, plugins: dict[str, list[ResourceValidationPlugin]]
) -> list[ValidationErrorRecord]:
    """Run all resource plugins on all resources of all classes of the given datapack."""
    records: list[ValidationErrorRecord] = []

    for class_name, class_resources in datapack.resources.items():
        for resource_id, resource in class_resources.items():
            for plugin in plugins.get(class_name, []):
                try:
                    plugin.validate(
                        resource_id=resource_id, resource=resource, datapack=datapack
                    )
                except ValidationPluginError as error:
                    records.append(
                        _plugin_error_to_record(error, subject_class=class_name)
                    )

    return records


class SchemaPackValidator:
    """A class for validating arbitrary datapacks against a specific schemapack."""

    def __init__(
        self,
        *,
        schemapack: SchemaPack,
        add_global_plugins: list[type[GlobalValidationPlugin]] | None = None,
        add_class_plugins: list[type[ClassValidationPlugin]] | None = None,
        add_resource_plugins: list[type[ResourceValidationPlugin]] | None = None,
    ):
        """Initialize with a specific schemapack.

        Args:
            schemapack:
                The schemapack to validate against.
            add_global_plugins:
                Global validation plugins to use in addition to the default ones.
            add_class_plugins:
                Class validation plugins to use in addition to the default ones.
            add_resource_plugins:
                Resource validation plugins to use in addition to the default ones.
        """
        self._global_plugins = _create_global_plugins(
            schemapack=schemapack,
            plugin_classes=DEFAULT_GLOBAL_PLUGIN_REGISTRY + (add_global_plugins or []),
        )
        self._class_plugins = _create_plugins_by_class(
            schemapack=schemapack,
            plugin_classes=DEFAULT_CLASS_PLUGIN_REGISTRY + (add_class_plugins or []),
        )
        self._resource_plugins = _create_plugins_by_class(
            schemapack=schemapack,
            plugin_classes=DEFAULT_RESOURCE_PLUGIN_REGISTRY
            + (add_resource_plugins or []),
        )

    def validate(self, *, datapack: DataPack):
        """Validates the given datapack against the configured schemapack. The execution
        order of validation plugins is as follows:
        (1) Run global validation plugins
        (2) Only upon success, execute class validation plugins.
        (3) Only upon success, run resource validation plugins.
        The reason for this conditional execution is that class plugins are assuming a
        intact state on the global level. Similarly, resource plugins are assuming an
        intact state on the class level.

        Raises:
            schemapack.exceptions.ValidationError: If any validation issues are found.
        """
        error_records = _run_global_plugins(
            datapack=datapack, plugins=self._global_plugins
        )

        if not error_records:
            error_records.extend(
                _run_class_plugins(datapack=datapack, plugins=self._class_plugins)
            )

            if not error_records:
                error_records.extend(
                    _run_resource_plugins(
                        datapack=datapack, plugins=self._resource_plugins
                    )
                )

        if error_records:
            raise ValidationError(records=error_records)



# Copyright 2021 - 2025 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Tests the main module."""

from pathlib import Path

import pytest

from schemapack import load_and_validate
from schemapack.exceptions import (
    BaseError,
    DataPackSpecError,
    ParsingError,
    ValidationError,
)
from tests.fixtures.examples import (
    INVALID_DATAPACK_PATHS,
    VALID_DATAPACK_PATHS,
    VALID_SCHEMAPACK_PATHS,
)


@pytest.mark.parametrize(
    "name, path", VALID_DATAPACK_PATHS.items(), ids=VALID_DATAPACK_PATHS
)
def test_load_and_validate_valid(name: str, path: Path):
    """Test load_and_validate function with valid schemapack and valid datapacks."""
    schemapack_name = name.split(".", 1)[0]
    schemapack_path = VALID_SCHEMAPACK_PATHS[schemapack_name]

    _ = load_and_validate(schemapack_path=schemapack_path, datapack_path=path)


@pytest.mark.parametrize(
    "name, path", INVALID_DATAPACK_PATHS.items(), ids=INVALID_DATAPACK_PATHS
)
def test_load_and_validate_invalid(name: str, path: Path):
    """Test load_and_validate function with valid schemapack but invalid datapacks."""
    schemapack_name, error_type = name.split(".", 2)[:2]
    schemapack_path = VALID_SCHEMAPACK_PATHS[schemapack_name]
    with pytest.raises(BaseError) as exception_info:
        _ = load_and_validate(schemapack_path=schemapack_path, datapack_path=path)
    # Either a DataPackSpecError or ValidationError is expected:
    if error_type == "DataPackSpecError":
        assert isinstance(exception_info.value, DataPackSpecError)
    elif error_type == "ParsingError":
        assert isinstance(exception_info.value, ParsingError)
    else:
        assert isinstance(exception_info.value, ValidationError)
        error_records = exception_info.value.records
        assert len(error_records) == 1
        assert error_records[0].type == error_type



# Copyright 2021 - 2025 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Default Validation Plugin Registry"""

from schemapack._internals.validation.base import (
    ClassValidationPlugin,
    GlobalValidationPlugin,
    ResourceValidationPlugin,
)
from schemapack._internals.validation.plugins import (
    ContentSchemaValidationPlugin,
    ExpectedRootValidationPlugin,
    MissingClassSlotValidationPlugin,
    MissingMandatoryOriginValidationPlugin,
    MissingMandatoryTargetValidationPlugin,
    MissingRelationValidationPlugin,
    MultipleTargetValidationPlugin,
    TargetOverlapValidationPlugin,
    UnexpectedRootValidationPlugin,
    UnknownClassSlotValidationPlugin,
    UnknownRelationValidationPlugin,
)

DEFAULT_GLOBAL_PLUGIN_REGISTRY: list[type[GlobalValidationPlugin]] = [
    MissingClassSlotValidationPlugin,
    UnknownClassSlotValidationPlugin,
    ExpectedRootValidationPlugin,
    UnexpectedRootValidationPlugin,
]
DEFAULT_CLASS_PLUGIN_REGISTRY: list[type[ClassValidationPlugin]] = [
    TargetOverlapValidationPlugin,
    MissingMandatoryOriginValidationPlugin,
]
DEFAULT_RESOURCE_PLUGIN_REGISTRY: list[type[ResourceValidationPlugin]] = [
    ContentSchemaValidationPlugin,
    MissingRelationValidationPlugin,
    UnknownRelationValidationPlugin,
    MultipleTargetValidationPlugin,
    MissingMandatoryTargetValidationPlugin,
]



# Copyright 2021 - 2025 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Collection of package-specific exceptions"""

from abc import ABC
from dataclasses import dataclass
from enum import Enum

import pydantic_core


class SpecType(str, Enum):
    """An enumeration of the types of specs."""

    SCHEMAPACK = "schemapack"
    DATAPACK = "datapack"


class BaseError(ABC, Exception):
    """Base class for all schemapack errors."""


class ParsingError(BaseError, ValueError):
    """Raised when parsing JSON or YAML data fails."""


class SpecError(BaseError, ValueError):
    """An error indicating that an instance was not valid against the specification."""

    def __init__(self, message: str, *, details: list[pydantic_core.ErrorDetails]):
        """Initiate a SpecError.

        Args:
            message: A human-readable message describing the error.
            details: Details on which fields of the instance are invalid.
        """
        super().__init__(message)
        self.message = message
        self.details = details


class SchemaPackSpecError(SpecError):
    """An error indicating that the provided data is not valid against the schemapack
    specification.
    """

    def __init__(self, message: str, *, details: list[pydantic_core.ErrorDetails]):
        """Initiate a SchemaPackSpecError.

        Args:
            message: A human-readable message describing the error.
            details: Details on which fields of the instance are invalid.
        """
        message = (
            "The provided object was not compatible with the schemapack"
            + f" specification: {message}"
        )
        super().__init__(message, details=details)


class DataPackSpecError(SpecError):
    """An error indicating that the provided data is not valid against the datapack
    specification (this is independent of validation done using a schemapack).
    """

    def __init__(self, message: str, *, details: list[pydantic_core.ErrorDetails]):
        """Initiate a SchemaPackSpecError.

        Args:
            message: A human-readable message describing the error.
            details: Details on which fields of the instance are invalid.
        """
        message = (
            "The provided object was not compatible with the datapack"
            + f" specification: {message}"
        )
        super().__init__(message, details=details)


class ValidationPluginError(BaseError, ValueError):
    """Raised by a ValidationPlugin."""

    def __init__(
        self, *, type_: str, message: str, details: dict[str, object] | None = None
    ):
        """Initiate a ValidationPluginError.

        Args:
            type: A preferably short type label.
            message: A human-readable message describing the error.
            details: A dictionary for transporting additional machine-readable details.
        """
        super().__init__(message)
        self.type_ = type_
        self.message = message
        self.details = details if details else {}


@dataclass
class ValidationErrorRecord:
    """A record of an Error occuring during validation on a specific context
    (e.g. a specific resource) regarding a specific validation aspect.

    Attributes:
        subject_class:
            If relevant, the name of the subject class as per the schemapack.
        subject_resource:
            If relevant, the ID or the subject resource of the subject class.
        type:
            A preferably short type label.
        message:
            A human-readable message describing the error.
        details:
            A dictionary for transporting additional machine-readable details.
    """

    subject_class: str | None
    subject_resource: str | None
    type: str
    message: str
    details: dict[str, object]


def _val_record_to_str(record: ValidationErrorRecord) -> str:
    """Translate a ValidationErrorRecords into a human-readable message
    to be displayed to the user on the terminal.
    """
    if record.subject_class:
        if record.subject_resource:
            context = f"resource '{record.subject_class}.{record.subject_resource}'"
        else:
            context = f"class '{record.subject_class}'"
    else:
        context = "global datapack"

    return (
        f"Error in {context}:"
        + f"\n\tType: {record.type}"
        + f"\n\tMessage: {record.message}"
    )


def _sorted_val_records(
    records: list[ValidationErrorRecord],
) -> list[ValidationErrorRecord]:
    """Sort a set of ValidationErrorRecords."""
    return sorted(records, key=lambda r: (r.subject_class, r.subject_resource, r.type))


class ValidationError(BaseError, ValueError):
    """A collection of ValidationErrorRecords raised when a datapack fails validation
    against a schemapack.
    """

    def __init__(self, records: list[ValidationErrorRecord]):
        """Initiate a ValidationError.

        Args:
            records: A list of ValidationErrorRecords.
        """
        if not records:
            raise ValueError("ValidationError must be raised with at least one record.")

        self.records = _sorted_val_records(records)
        n_records = len(self.records)
        record_messages = "\n".join(_val_record_to_str(r) for r in self.records)
        message = f"Validation failed with {n_records} issue(s):\n{record_messages}"

        super().__init__(message)


class ValidationAssumptionError(BaseError, RuntimeError):
    """This is raised when a unit of code assumes to work with a datapack that has
    already been validated against a specific schemapack, however, it became apparent
    that this assumption was not correct. The error should not contain the details on
    which aspect is invalid. These details can be retrieved by performing a validation
    and receiving a ValidationError.
    """

    def __init__(self, *, context: str):
        """Initiate a ValidationAssumptionError.

        Args:
            context: Some context to be included in the error message.
        """
        message = (
            "It was assumed that the provided datapack has already been validated"
            + " against the provided schemapack, however, this assumption failed in"
            + f" the context of {context}. Please validate the datapack against the"
            + " schemapack to get further details."
        )
        super().__init__(message)
        self.message = message


class CircularRelationError(BaseError, ValueError):
    """Raised when a circular relation between resources is detected, but the requested
    operation cannot be performed on datapacks with circular relations.
    """


class ClassNotFoundError(BaseError, KeyError):
    """Raised when a class was not found in a schemapack or datapack."""

    def __init__(self, *, class_name: str, spec_type: SpecType):
        """Initiate a ClassNotFoundError.

        Args:
            class_name:
                The name of the class that was not found.
            spec_type:
                The type of spec that the class was not found in.
        """
        message = f"Class '{class_name}' not found in the provided {spec_type.value}."
        super().__init__(message)
        self.class_name = class_name


class ResourceNotFoundError(BaseError, KeyError):
    """Raised when a resource was not found in a datapack."""

    def __init__(self, *, class_name: str, resource_id: str):
        """Initiate a ResourceNotFoundError.

        Args:
            class_name:
                The name of the class of the resource that was not found.
            resource_id:
                The ID of the resource that was not found.
        """
        message = (
            f"Resource of class '{class_name}' with id '{resource_id}' not found in"
            + " the provided datapack."
        )
        super().__init__(message)
        self.class_name = class_name
        self.resource_id = resource_id


class InvalidEmbeddingProfileError(BaseError, ValueError):
    """Raised when an embedding profile is invalid. A profile is invalid if the value
    of any key is neither a boolean nor a nested dictionary following the same rules
    recursively (i.e., the nested dictionary values must also be booleans or
    nested dictionaries).

    Invalid examples:

    A value that is neither a boolean nor a dictionary:
        {"some_relation": "this is not a boolean or a dictionary"}
    A nested dictionary containing invalid values (e.g., a string instead of a boolean):
        {"some_relation": {"nested_relation": "this is not a boolean"}}

    Valid examples:
        {"some_relation": True}
        {"some_relation": {"nested_relation": False}}
        {"some_relation": {"nested_relation": {"another_nested_relation": True}}}
    """



# Copyright 2021 - 2025 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for implementing a CLI"""

import rich.console
import rich.panel

EXCEPTION_CONSOLE = rich.console.Console(stderr=True, style="red")
FINAL_SUCCESS_CONSOLE = rich.console.Console(style="green bold", stderr=True)
FINAL_FAILURE_CONSOLE = rich.console.Console(style="red bold", stderr=True)
OUTPUT_CONSOLE = rich.console.Console(soft_wrap=True, markup=False)


def print_exception(exception: Exception, exception_name: str = "Exception"):
    """Print an exception."""
    EXCEPTION_CONSOLE.print(rich.panel.Panel(str(exception), title=exception_name))


def print_final_success(
    message: str,
):
    """Print a final message indicating success."""
    FINAL_SUCCESS_CONSOLE.print(message)


def print_final_failure(
    message: str,
):
    """Print a final message indicating failure."""
    FINAL_FAILURE_CONSOLE.print(message)


def print_output(
    message: str,
):
    """Print primary output of the command."""
    OUTPUT_CONSOLE.print(message)



# Copyright 2021 - 2025 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""High-level convenience functions.

Warning: This is an internal part of the library and might change without notice.
"""

from pathlib import Path

from schemapack._internals.load import load_datapack, load_schemapack
from schemapack._internals.validation import SchemaPackValidator
from schemapack.spec.datapack import DataPack
from schemapack.spec.schemapack import SchemaPack


def load_and_validate(
    *, schemapack_path: Path, datapack_path: Path
) -> tuple[SchemaPack, DataPack]:
    """Load and validate a datapack with respect to a schemapack using the default
    validation plugins.

    Raises:
        schemapack.exceptions.SchemaPackFormatError:
            If the schemapack has an invalid format.
        schemapack.exceptions.DataPackFormatError:
            If the datapack has an invalid format.
        schemapack.exceptions.ValidationError:
            If the datapack is not valid against the schemapack.
    """
    schemapack = load_schemapack(schemapack_path)
    datapack = load_datapack(datapack_path)
    validator = SchemaPackValidator(schemapack=schemapack)
    validator.validate(datapack=datapack)
    return schemapack, datapack



# Copyright 2021 - 2025 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test the cli.

Please note, these tests will just shallowly test that the cli commands can be called
and produce the expected output. However, the output is not deeply investigated and
only a single senario is tester for each command. More deeper testing is done on the
python API (to avoid redundancy of tests).
"""

from pathlib import Path

import pytest
import ruamel.yaml
from typer.testing import CliRunner

from schemapack import __version__ as schemapack_version
from schemapack._internals.cli import cli
from schemapack._internals.utils import read_json_or_yaml_mapping
from schemapack.cli import exit_codes
from tests.fixtures.examples import (
    ERD_PATHS,
    INVALID_DATAPACK_PATHS,
    INVALID_SCHEMAPACK_PATHS,
    VALID_DATAPACK_PATHS,
    VALID_SCHEMAPACK_PATHS,
)
from tests.fixtures.utils import (
    assert_formatted_string,
    loads_json_or_yaml_mapping,
    strip_ansi_escape_sequences,
)

yaml = ruamel.yaml.YAML(typ="rt")
runner = CliRunner()


def test_version():
    """Test the version command."""
    result = runner.invoke(cli, ["--version"])
    assert result.exit_code == exit_codes.SUCCESS == 0
    assert strip_ansi_escape_sequences(result.stdout).strip() == str(schemapack_version)


def generate_validate_command(
    *, schemapack: Path, datapack: Path, abbreviate: bool
) -> list[str]:
    """Generate a command line to validate a datapack against a schemapack."""
    return [
        "validate",
        "-s" if abbreviate else "--schemapack",
        str(schemapack),
        "-d" if abbreviate else "--datapack",
        str(datapack),
    ]


@pytest.mark.parametrize(
    "abbreviate", [True, False], ids=["abbreviate", "no_abbreviate"]
)
def test_validate_valid(abbreviate: bool):
    """Test the validate command with a valid datapack."""
    schemapack = VALID_SCHEMAPACK_PATHS["simple_relations"]
    datapack = VALID_DATAPACK_PATHS["simple_relations.simple_resources"]

    result = runner.invoke(
        cli,
        generate_validate_command(
            schemapack=schemapack, datapack=datapack, abbreviate=abbreviate
        ),
    )
    assert result.exit_code == exit_codes.SUCCESS == 0


@pytest.mark.parametrize(
    "abbreviate", [True, False], ids=["abbreviate", "no_abbreviate"]
)
def test_validate_invalid(abbreviate: bool):
    """Test the validate command with an invalid datapack."""
    schemapack = VALID_SCHEMAPACK_PATHS["simple_relations"]
    datapack = INVALID_DATAPACK_PATHS[
        "simple_relations.ContentValidationError.missing_property"
    ]

    result = runner.invoke(
        cli,
        generate_validate_command(
            schemapack=schemapack, datapack=datapack, abbreviate=abbreviate
        ),
    )
    assert result.exit_code == exit_codes.VALIDATION_ERROR != 0
    assert "ContentValidationError" in result.stderr


@pytest.mark.parametrize(
    "abbreviate", [True, False], ids=["abbreviate", "no_abbreviate"]
)
def test_validate_datapack_spec_error(abbreviate: bool):
    """Test the validate command with a datapack that does not comply with the specs."""
    schemapack = VALID_SCHEMAPACK_PATHS["simple_relations"]
    datapack = INVALID_DATAPACK_PATHS[
        "simple_relations.DataPackSpecError.DuplicateTargetIdError"
    ]

    result = runner.invoke(
        cli,
        generate_validate_command(
            schemapack=schemapack, datapack=datapack, abbreviate=abbreviate
        ),
    )
    assert result.exit_code == exit_codes.DATAPACK_SPEC_ERROR != 0
    assert "DataPackSpecError" in result.stderr


@pytest.mark.parametrize(
    "abbreviate", [True, False], ids=["abbreviate", "no_abbreviate"]
)
def test_validate_schemapack_spec_error(abbreviate: bool):
    """Test the validate command with a schemapack that does not comply with the specs."""
    schemapack = INVALID_SCHEMAPACK_PATHS["ContentSchemaNotFoundError"]
    datapack = VALID_DATAPACK_PATHS["simple_relations.simple_resources"]

    result = runner.invoke(
        cli,
        generate_validate_command(
            schemapack=schemapack, datapack=datapack, abbreviate=abbreviate
        ),
    )
    assert result.exit_code == exit_codes.SCHEMAPACK_SPEC_ERROR != 0
    assert "SchemaPackSpecError" in result.stderr


def test_check_schemapack_complies():
    """Test the check-schemapack command with a complies document."""
    schemapack = VALID_SCHEMAPACK_PATHS["simple_relations"]

    result = runner.invoke(cli, ["check-schemapack", str(schemapack)])
    assert result.exit_code == exit_codes.SUCCESS == 0


def test_check_schemapack_not_complies():
    """Test the check-schemapack command with a non compliant document."""
    schemapack = INVALID_SCHEMAPACK_PATHS["ContentSchemaNotFoundError"]

    result = runner.invoke(cli, ["check-schemapack", str(schemapack)])
    assert result.exit_code == exit_codes.SCHEMAPACK_SPEC_ERROR != 0
    assert "SchemaPackSpecError" in result.stderr


def test_check_datapack_complies():
    """Test the check-datapack command with a complies document."""
    datapack = VALID_DATAPACK_PATHS["simple_relations.simple_resources"]

    result = runner.invoke(cli, ["check-datapack", str(datapack)])
    assert result.exit_code == exit_codes.SUCCESS == 0


def test_check_datapack_not_complies():
    """Test the check-datapack command with a non compliant document."""
    datapack = INVALID_DATAPACK_PATHS[
        "simple_relations.DataPackSpecError.DuplicateTargetIdError"
    ]

    result = runner.invoke(cli, ["check-datapack", str(datapack)])
    assert result.exit_code == exit_codes.DATAPACK_SPEC_ERROR != 0
    assert "DataPackSpecError" in result.stderr


@pytest.mark.parametrize(
    "json_format, abbreviate",
    [(False, False), (True, False), (True, True)],
    ids=["yaml", "json", "json_abbrev"],
)
def test_condense_schemapack(json_format: bool, abbreviate: bool):
    """Test the condense-schemapack command."""
    schemapack_path = VALID_SCHEMAPACK_PATHS["simple_relations"]
    expected_dict = read_json_or_yaml_mapping(
        VALID_SCHEMAPACK_PATHS["simple_relations_condensed"]
    )

    command = ["condense-schemapack", str(schemapack_path)]
    if json_format:
        command.append("-j" if abbreviate else "--json")
    result = runner.invoke(cli, command)
    assert result.exit_code == exit_codes.SUCCESS == 0

    observed_str = strip_ansi_escape_sequences(result.stdout)
    assert_formatted_string(observed_str, json_format=json_format)

    observed_dict = loads_json_or_yaml_mapping(observed_str)
    assert observed_dict == expected_dict


@pytest.mark.parametrize(
    "json_format, abbreviate",
    [(False, False), (True, False), (True, True)],
    ids=["yaml", "json", "json_abbrev"],
)
def test_isolate_resource(json_format: bool, abbreviate: bool):
    """Test the isolate_resource command."""
    schemapack_path = VALID_SCHEMAPACK_PATHS["simple_relations"]
    datapack_path = VALID_DATAPACK_PATHS["simple_relations.simple_resources"]
    expected_dict = read_json_or_yaml_mapping(
        VALID_DATAPACK_PATHS["simple_relations_rooted.rooted_simple_resources"]
    )
    class_name = "Dataset"
    resource_id = "example_dataset_1"

    command = [
        "isolate-resource",
        "-s" if abbreviate else "--schemapack",
        str(schemapack_path),
        "-d" if abbreviate else "--datapack",
        str(datapack_path),
        "-c" if abbreviate else "--class-name",
        class_name,
        "-r" if abbreviate else "--resource-id",
        resource_id,
    ]
    if json_format:
        command.append("-j" if abbreviate else "--json")
    result = runner.invoke(cli, command)
    assert result.exit_code == exit_codes.SUCCESS == 0

    observed_str = strip_ansi_escape_sequences(result.output)
    assert_formatted_string(observed_str, json_format=json_format)

    observed_dict = loads_json_or_yaml_mapping(observed_str)
    assert observed_dict == expected_dict


def test_isolate_resource_non_existing_class():
    """Test the isolate_resource command with a non-existing class."""
    schemapack_path = VALID_SCHEMAPACK_PATHS["simple_relations"]
    datapack_path = VALID_DATAPACK_PATHS["simple_relations.simple_resources"]
    class_name = "NonExistingClass"
    resource_id = "example_dataset_1"

    command = [
        "isolate-resource",
        "--schemapack",
        str(schemapack_path),
        "--datapack",
        str(datapack_path),
        "--class-name",
        class_name,
        "--resource-id",
        resource_id,
    ]
    result = runner.invoke(cli, command)
    assert result.exit_code == exit_codes.CLASS_NOT_FOUND_ERROR != 0
    assert "ClassNotFoundError" in result.stderr


def test_isolate_resource_non_existing_resource():
    """Test the isolate_resource command with a non-existing resource."""
    schemapack_path = VALID_SCHEMAPACK_PATHS["simple_relations"]
    datapack_path = VALID_DATAPACK_PATHS["simple_relations.simple_resources"]
    class_name = "Dataset"
    resource_id = "non_existing_resource"

    command = [
        "isolate-resource",
        "--schemapack",
        str(schemapack_path),
        "--datapack",
        str(datapack_path),
        "--class-name",
        class_name,
        "--resource-id",
        resource_id,
    ]
    result = runner.invoke(cli, command)
    assert result.exit_code == exit_codes.RESOURCE_NOT_FOUND_ERROR != 0
    assert "ResourceNotFoundError" in result.stderr


@pytest.mark.parametrize(
    "json_format, abbreviate",
    [(False, False), (True, False), (True, True)],
    ids=["yaml", "json", "json_abbrev"],
)
def test_isolate_class(json_format: bool, abbreviate: bool):
    """Test the isolate_class command."""
    schemapack_path = VALID_SCHEMAPACK_PATHS["simple_relations"]
    expected_dict = read_json_or_yaml_mapping(
        VALID_SCHEMAPACK_PATHS["simple_relations_rooted_condensed"]
    )
    class_name = "Dataset"

    command = [
        "isolate-class",
        "-s" if abbreviate else "--schemapack",
        str(schemapack_path),
        "-c" if abbreviate else "--class-name",
        class_name,
    ]
    if json_format:
        command.append("-j" if abbreviate else "--json")
    result = runner.invoke(cli, command)
    assert result.exit_code == exit_codes.SUCCESS == 0

    observed_str = strip_ansi_escape_sequences(result.output)
    assert_formatted_string(observed_str, json_format=json_format)

    observed_dict = loads_json_or_yaml_mapping(observed_str)
    assert observed_dict == expected_dict


def test_isolate_class_non_existing():
    """Test the isolate_class command with a non-existing class."""
    schemapack_path = VALID_SCHEMAPACK_PATHS["simple_relations"]
    class_name = "NonExistingClass"

    command = [
        "isolate-class",
        "--schemapack",
        str(schemapack_path),
        "--class-name",
        class_name,
    ]
    result = runner.invoke(cli, command)
    assert result.exit_code == exit_codes.CLASS_NOT_FOUND_ERROR != 0
    assert "ClassNotFoundError" in result.stderr


@pytest.mark.parametrize(
    "props,example_suffix",
    [
        ([], "_wo_props"),
        (["-c"], "_w_props"),
        (["--content-properties"], "_w_props"),
    ],
    ids=("no_content_props", "content_props_abbrev", "content_props"),
)
def test_export_mermaid(tmp_path, props: list[str], example_suffix: str):
    """Test the export-mermaid command."""
    example = "comprehensive_cardinalities_and_types"
    schemapack = VALID_SCHEMAPACK_PATHS[example]
    erd = ERD_PATHS[example + example_suffix]

    args = [
        "export-mermaid",
        str(schemapack),
        *props,
    ]
    result = runner.invoke(cli, args)
    assert result.exit_code == exit_codes.SUCCESS == 0
    assert strip_ansi_escape_sequences(result.output) == erd.read_text()



# Copyright 2021 - 2025 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A specification (plus tooling) for describing linked data models based on JSON
schema.
"""

from importlib.metadata import version

from ._internals.dump import dump_schemapack, dumps_datapack, dumps_schemapack
from ._internals.erd import export_mermaid
from ._internals.isolate import isolate, isolate_class, isolate_resource
from ._internals.load import load_datapack, load_schemapack
from ._internals.main import load_and_validate
from ._internals.normalize import denormalize
from ._internals.validation import SchemaPackValidator

__all__ = [
    "SchemaPackValidator",
    "denormalize",
    "dump_schemapack",
    "dumps_datapack",
    "dumps_schemapack",
    "export_mermaid",
    "isolate",
    "isolate_class",
    "isolate_resource",
    "load_and_validate",
    "load_datapack",
    "load_schemapack",
]

__version__ = version(__package__)



# Copyright 2021 - 2025 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Logic for exporting schemapacks in mermaid format."""

from collections.abc import Mapping, Sequence
from typing import Any

from schemapack._internals.spec.schemapack import ClassDefinition, ClassRelation
from schemapack.spec.schemapack import SchemaPack


def get_property_type(json_schema_props: Mapping[str, Any], prop_name: str) -> str:
    """Get the type of a property from a JSON schema properties map. If no type is
    specified, 'object' is assumed. If the property has an enum key, 'enum' is returned.

    Args:
        json_schema_props: The properties of the class.
        prop_name: The name of the property to get the type of.

    Returns:
        The type of the property.
    """
    prop = json_schema_props[prop_name]
    if isinstance(prop, Mapping):
        type = prop.get("type", "object") if not prop.get("enum") else "enum"
        if type == "array":
            element_type = prop.get("items", {}).get("type", "object")
            return f"array[{element_type}]"
        return type
    raise ValueError(
        f"Invalid JSON schema. Expected property {prop_name} to be a mapping."
    )


def export_class_entity(
    class_name: str, class_def: ClassDefinition, content_properties: bool
) -> str:
    """
    Export a class entity in mermaid format.

    Args:
        class_name: The name of the class to export.
        class_def: The definition of the class to export.
        show_properties: Whether to include the properties of the class in the diagram.

    Returns:
        A string representing the class entity in mermaid format.
    """
    if not content_properties:
        return f"{class_name} {{}}"

    json_schema_obj_props = class_def.content.get("properties", {})
    additional_properties = class_def.content.get("additionalProperties", False)
    json_schema_obj_reqs = class_def.content.get("required", [])

    if not isinstance(json_schema_obj_reqs, Sequence):
        raise ValueError("Invalid JSON schema. Expected 'required' to be a list.")

    fields = [
        f"  {get_property_type(json_schema_props=json_schema_obj_props, prop_name=field)}"
        + f" {field} "
        + ('"req"' if field in json_schema_obj_reqs else '"opt"')
        for field in json_schema_obj_props
    ]

    if additional_properties:
        fields.append('  * * ""')

    fields_str = "\n" + "\n".join(fields) + "\n" if fields else ""

    return f"{class_name} {{{fields_str}}}"


def export_class_relation(
    class_name: str, property_name: str, relation: ClassRelation
) -> str:
    """Export a relation between two classes in mermaid format.

    Args:
        class_name: The name of the class that contains the relation.
        property_name: The name of the relation.
        relation: The relation definition.

    Returns:
        A string representing the relation between the two classes in mermaid format.
    """
    origin_multiple = "}" if relation.multiple.origin else "|"
    target_multiple = "{" if relation.multiple.target else "|"
    origin_mandatory = "|" if relation.mandatory.origin else "o"
    target_mandatory = "|" if relation.mandatory.target else "o"
    return (
        f"{class_name} "
        f"{origin_multiple}{origin_mandatory}--"
        + f"{target_mandatory}{target_multiple}"
        + f' {relation.targetClass} : "{class_name}.{property_name}"'
    )


def export_class(
    class_name: str, class_def: ClassDefinition, content_properties: bool
) -> str:
    """Export a class in mermaid format.

    Args:
        class_name: The name of the class to export.
        class_def: The definition of the class to export.
        show_properties: Whether to include the properties of the class in the diagram.

    Returns:
        A string representing the class in mermaid format.
    """
    class_str = export_class_entity(
        class_name=class_name,
        class_def=class_def,
        content_properties=content_properties,
    )

    relations_str = (
        "\n\n"
        + "\n".join(
            export_class_relation(class_name, prop_name, relation)
            for prop_name, relation in class_def.relations.items()
        )
        if class_def.relations
        else ""
    )

    return class_str + relations_str


def export_mermaid(schemapack: SchemaPack, content_properties: bool = True) -> str:
    """
    Export a SchemaPack in mermaid format.

    Args:
        schemapack: The SchemaPack to export.
        show_properties: Whether to include the properties of the classes in the diagram.

    Returns:
        A string representing the SchemaPack in mermaid format.
    """
    erd_str = "erDiagram\n\n" + "\n\n".join(
        export_class(
            class_name=class_name,
            class_def=class_def,
            content_properties=content_properties,
        )
        for class_name, class_def in schemapack.classes.items()
    )

    return erd_str



# Copyright 2021 - 2025 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Base classes for defining validation plugins."""

from abc import ABC, abstractmethod
from collections.abc import Mapping

from schemapack.spec.custom_types import ResourceId
from schemapack.spec.datapack import DataPack, Resource
from schemapack.spec.schemapack import ClassDefinition, SchemaPack


class GlobalValidationPlugin(ABC):
    """Abstract class for a plugin that performs validation of a datapack with respect
    to the entire schemapack.

    Please note:
    A ValidationPlugin should always just check for one single aspect of validation
    and should not perform all checks possible with a schemapack.
    The "global" scope only means that objects implementing this protocol do not focus
    on one specific datapack class or resource but rather on a validation
    aspect that applies to the entire datapack.
    """

    @staticmethod
    @abstractmethod
    def does_apply(*, schemapack: SchemaPack) -> bool:
        """A classmethod to check whether this validation plugin is relevant for the
        given schemapack.

        Returns: True if this plugin is relevant for the given schemapack.
        """
        raise NotImplementedError

    @abstractmethod
    def __init__(self, *, schemapack: SchemaPack):
        """This plugin is configured with the entire schemapack."""
        raise NotImplementedError

    @abstractmethod
    def validate(self, *, datapack: DataPack):
        """Validate the entire datapack.

        Raises:
            schemapack.exceptions.ValidationPluginError: If validation fails.
        """
        raise NotImplementedError


class ClassValidationPlugin(ABC):
    """Abstract class for defining class-scoped validation plugin. Instances
    of implemeting concrete classes perform validation with respect to one
    specific schemapack class.
    """

    @staticmethod
    @abstractmethod
    def does_apply(*, class_: ClassDefinition) -> bool:
        """A classmethod to check whether this validation plugin is relevant for the
        given class definition.

        Returns: True if this plugin is relevant for the given class definition.
        """
        raise NotImplementedError

    @abstractmethod
    def __init__(self, *, class_: ClassDefinition):
        """This plugin is configured with one specific class definition of a schemapack."""
        raise NotImplementedError

    @abstractmethod
    def validate(
        self, *, class_resources: Mapping[ResourceId, Resource], datapack: DataPack
    ):
        """Validate all resources of a specific class. The entire datapack is provided
        for resolving relations to other classes.

        Raises:
            schemapack.exceptions.ValidationPluginError: If validation fails.
        """
        raise NotImplementedError


class ResourceValidationPlugin(ABC):
    """Abstract class for defining resource-scoped validation plugin. Implementing
    objects perform validation with respect to one specific resource of a schemapack
    class.
    """

    @staticmethod
    @abstractmethod
    def does_apply(*, class_: ClassDefinition) -> bool:
        """A classmethod to check whether this validation plugin is relevant for the
        given class definition.

        Returns: True if this plugin is relevant for the given class definition.
        """
        raise NotImplementedError

    @abstractmethod
    def __init__(self, *, class_: ClassDefinition):
        """This plugin is configured with one specific class definition of a schemapack."""
        raise NotImplementedError

    @abstractmethod
    def validate(
        self, *, resource: Resource, resource_id: ResourceId, datapack: DataPack
    ) -> None:
        """Validates a specific resource of the defined class. The entire datapack is
        provided for resolving relations to resources of other classes.

        Raises:
            schemapack.exceptions.ValidationPluginError: If validation fails.
        """
        raise NotImplementedError



# Copyright 2021 - 2025 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests the load module."""

from contextlib import nullcontext
from pathlib import Path

import pytest

from schemapack import load_datapack, load_schemapack
from schemapack._internals.spec.schemapack import ClassDefinition
from schemapack.exceptions import DataPackSpecError, ParsingError, SchemaPackSpecError
from tests.fixtures.examples import (
    INVALID_DATAPACK_PATHS,
    INVALID_SCHEMAPACK_PATHS,
    VALID_DATAPACK_PATHS,
    VALID_SCHEMAPACK_PATHS,
)
from tests.fixtures.test_content_schema_objects import (
    DATASET_CONTENT,
    DATASET_DESCRIPTION,
    DATASET_ID,
    DATASET_RELATIONS,
)


@pytest.mark.parametrize(
    "path", VALID_SCHEMAPACK_PATHS.values(), ids=VALID_SCHEMAPACK_PATHS
)
def test_load_schemapack_valid(path: Path):
    """Test loading valid schemapacks."""
    _ = load_schemapack(path)


@pytest.mark.parametrize(
    "name, path", INVALID_SCHEMAPACK_PATHS.items(), ids=INVALID_SCHEMAPACK_PATHS
)
def test_load_schemapack_invalid(name: str, path: Path):
    """Test loading invalid schemapacks."""
    error_type = name.split(".")[0]

    with pytest.raises(SchemaPackSpecError) as exception_info:
        _ = load_schemapack(path)

    error_details = exception_info.value.details
    assert len(error_details) == 1

    assert error_details[0]["type"] == error_type


@pytest.mark.parametrize(
    "path", VALID_DATAPACK_PATHS.values(), ids=VALID_DATAPACK_PATHS
)
def test_load_datapack_valid(path: Path):
    """Test loading valid datapacks."""
    _ = load_datapack(path)


@pytest.mark.parametrize(
    "name, path", INVALID_DATAPACK_PATHS.items(), ids=INVALID_DATAPACK_PATHS
)
def test_load_datapack_invalid(name: str, path: Path):
    """Test loading invalid datapacks."""
    error_type = name.split(".")[1]
    expected_error = (
        DataPackSpecError
        if error_type == "DataPackSpecError"
        else ParsingError
        if error_type == "ParsingError"
        else None
    )

    with pytest.raises(expected_error) if expected_error else nullcontext():
        _ = load_datapack(path)


def test_load_content_schema():
    """Test creating a ClassDefinition object with a content schema given as a FrozenDict.
    This ensures the jsonschema schema validation works for FrozenDict.
    """
    _ = ClassDefinition(
        id=DATASET_ID,
        description=DATASET_DESCRIPTION,
        content=DATASET_CONTENT,
        relations=DATASET_RELATIONS,
    )



# Copyright 2021 - 2025 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""A validation plugin."""

from schemapack._internals.validation.base import GlobalValidationPlugin
from schemapack.exceptions import ValidationPluginError
from schemapack.spec.datapack import DataPack
from schemapack.spec.schemapack import SchemaPack


class MissingClassSlotValidationPlugin(GlobalValidationPlugin):
    """A global-scoped validation plugin validating that a datapack has a slot for
    each class defined in the provided schemapack.
    """

    @staticmethod
    def does_apply(*, schemapack: SchemaPack) -> bool:
        """A classmethod to check whether this validation plugin is relevant for the
        given schemapack.

        Returns: True if this plugin is relevant for the given class definition.
        """
        return True

    def __init__(self, *, schemapack: SchemaPack):
        """This plugin is configured with the entire schemapack."""
        self._classes = schemapack.classes.keys()

    def validate(self, *, datapack: DataPack):
        """Validate the entire datapack.

        Raises:
            schemapack.exceptions.ValidationPluginError: If validation fails.
        """
        missing_classes = [
            class_ for class_ in self._classes if class_ not in datapack.resources
        ]

        if missing_classes:
            raise ValidationPluginError(
                type_="MissingClassSlotError",
                message=("Missing slot(s) for class(es):" + ", ".join(missing_classes)),
                details={
                    "missing_classes": missing_classes,
                },
            )
